﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

public class Sharp2Advanced : MonoBehaviour
{
    void Start()
    {
        // ********** 高速な連結 **********

        // 文字列の連結を何度も行う場合は、StringBuilderを使う

        var sb = new StringBuilder();
        sb.Append("私は");
        sb.Append("文字です");
        sb.Append("よ！");

        // += で連結する度にメモリ確保が発生する遅い


        // ********** メモリ軽減な結合 **********

        // 文字列以外は混ざる場合の連結は、ToString() で全て文字列型に揃える

        var len = sb.Length;
        var good = sb.ToString() + "の文字数: " + len.ToString();

        // len.ToString() の ToString() が無くても結果は同じだが
        // object + object + object の連結になり、ボクシングでメモリ確保が発生する

        var bad = sb.ToString() + "の文字数: " + len;

        bad.Log();

        // コード
        // https://sharplab.io/#v2:EYLgZgpghgLgrgJwgZwLQGMD2A7ZMFQCW2MyANDCIQDZkAmIA1AD4ACADAASsCMA3AFgAUKwDM3AEycAwpwDewzku7jWAFk4BZABQBKeYuVGAblASdkwTgF5O2CAHduPAHQAVCAA8YLgMr5iAHMAITgaOggEPUEhAEhY03NkGwtgd0x/BCC9TkZOACIofNzUlwAZCGxAmAALGKMAX2EGoA==

        // 資料
        // https://www.slideshare.net/neuecc/cedec-2018-c-c/83


        // ********** 正確で高速な判定 **********

        // StringComparisonでOrdinalを指定する

        var b1 = good.Contains("文字");  // これは高速かつ正確

        var b2 = good.IndexOf("文字") >= 0;  // これは遅い
        var b3 = good.IndexOf("文字", StringComparison.Ordinal) >= 0;  // これは高速かつ正確

        // 正確とはバイナリ比較であり、カルチャ(言語環境)に依存しない

        // .NET の文字列を使用するためのベスト プラクティス | Microsoft Docs
        // https://docs.microsoft.com/ja-jp/dotnet/standard/base-types/best-practices-strings#common-string-comparison-methods-in-net


        // ********** 高速な列挙 ********** 

        // ----- bool[] -----

        // T[] なら気にせず foreach を使う

        var array = new[] { b1, b2, b3 };

        // 最速
        for (int i = 0; i < array.Length; i++)
        {
            Dummy.Log(array[i]);
        }

        // 最速2 (上のforに変換されて実行される)
        foreach (var b in array)
        {
            Dummy.Log(b);
        }


        // ----- List<bool> -----

        // List<T> で高速にするなら for を使う
        // T[] で済む場合は List<T> は使用しない (ToList() ではなく ToArray() にする)

        var list = array.ToList();

        // 最速
        for (int i = 0; i < list.Count; i++)
        {
            Dummy.Log(list[i]);
        }

        // 遅くなる (安全)
        foreach (var b in list)
        {
            Dummy.Log(b);

            // foreachの場合は、列挙中に変更すると例外発生
            // list.RemoveAt(0);

            // つまり、列挙中にコレクションが変更されないことが保証される
        }


        // ----- IEnumerable<bool> -----

        // 生産性を犠牲にする必要はないので、過度な最適化はしなくて良い

        // 遅くなる
        IEnumerable<bool> enumerable = array;

        foreach (var b in enumerable)
        {
            Dummy.Log(b);
        }


        // ********** 配列の比較 **********

        var array1 = new[] { 1, 2, 3 };
        var array2 = new[] { 1, 2, 3 };
        var ref1 = array1;

        // ----- 配列のインスタンスの比較 (中身ではなく配列という入れ物を比較) -----

        // 中身は同じでもインスタンスが違う
        b1 = array1 == array2;
        b2 = array1.Equals(2);
        b3 = object.ReferenceEquals(array1, array2);

        // 同じインスタンス (当然中身も同じ)
        b1 = array1 == ref1;
        b2 = array1.Equals(ref1);
        b3 = object.ReferenceEquals(array1, ref1);


        // ----- 比較方法 -----

        // == → 同じ型同士の比較
        // Equals() → 相手の型が異なる(不明の)場合の比較

        double d = 0;
        int val = 0;

        b1 = d == val;       // double と double の比較になる → true
        b2 = d.Equals(val);  // double と double の比較になる → true

        b1 = val == d;       // double と double の比較       → true
        b2 = val.Equals(d);  // double と object(int) の比較  → false!!!

        // int は double に暗黙的な変換が可能 (doubleの方が扱える範囲が広いので)

        // (参考) class, struct, 匿名型, タプルなど型が変わると比較結果も異なる (今後の回で解説)


        // ----- 件数と中身の一致比較 (中身のみで比較) -----

        // 中身はどれも一緒
        b1 = array1.SequenceEqual(array2);
        b2 = array1.SequenceEqual(ref1);


        // ----- 常にインスタンスで比較 -----

        // ピュアなC#クラスなら、object.ReferenceEqualsで比較して良い
        // GameObjectクラスなら、== で比較した方が安全
        b1 = object.ReferenceEquals(array1, array2);

        // object.ReferenceEquals(someObject, null) の方が someObject == null より早い…が、盲目的に使うのは危険 - テラシュールブログ
        // http://tsubakit1.hateblo.jp/entry/2015/05/28/014440
    }
}